package communication.handler;

import java.io.IOException;
import java.util.Enumeration;
import java.util.Vector;

import communication.protocol.ChunkReader;
import communication.protocol.ChunkWriter;
import communication.protocol.ProtocolListener;
import communication.transport.ConnectionListener;
import communication.transport.ConnectionProvider;

/**
 * Manages basic protocol capability, such as connection/disconnection and
 * capabilities configuration
 * 
 * @author Paolo
 */
public class ProtocolHandler implements Runnable, ProtocolListener
{
	public static boolean SUPPORT_THREAD_INTERRUPT = true;

	public static boolean SUPPORT_THREAD_DAEMON = true;

	private CapabilityHandler[] protocolHandlers;

	private ConnectionProvider provider;

	private ChunkReader reader;

	private ChunkWriter writer;

	private int capabilityMask;

	private Thread readerThread;

	private Vector connectionListeners;

	private int maxSpeed;

	private volatile boolean guard;

	private boolean connected;

	private class Event
	{
		public boolean arrived = false;
	}

	private Event tester;

	private Event setupper;

	/**
	 * Create a protocol handler using the specified connection provider
	 * 
	 * @param provider
	 *            A new connection provider
	 * @param maxBitRate
	 *            Maximum available bitrate (indication only)
	 */
	public ProtocolHandler(ConnectionProvider provider, int maxBitRate)
	{
		this.provider = provider;
		tester = new Event();
		setupper = new Event();
		writer = new ChunkWriter();
		reader = new ChunkReader();
		reader.setProtocolListener(this);
		readerThread = new Thread(this);
		protocolHandlers = new CapabilityHandler[8];
		connectionListeners = new Vector(1);
		maxSpeed = maxBitRate;
	}

	/**
	 * Clean up capability handlers.
	 */
	private void disposeHandlers()
	{
		// dispose capability handlers
		for (int i = 0; i < protocolHandlers.length; ++i)
		{
			if (protocolHandlers[i] != null)
			{
				protocolHandlers[i].disposeHandler();
				protocolHandlers[i] = null;
			}
		}
		connectionListeners.removeAllElements();
	}

	/**
	 * Notifies capability handlers to cancel pending operations.
	 */
	private void cancelHandlers()
	{
		// cancel capability handlers
		for (int i = 0; i < protocolHandlers.length; ++i)
		{
			if (protocolHandlers[i] != null)
				protocolHandlers[i].cancelHandler();
		}
	}

	/**
	 * Adds a new protocol capability handler. The capability mask is updated
	 * and the handler is registered as the appropriate protocol listener.
	 * 
	 * @param handler
	 *            The capability handler to use.
	 */
	public synchronized void addCapabilityHandler(CapabilityHandler handler)
	{
		int mask = handler.initializeHandler(reader, writer);
		for (int i = 0; i < protocolHandlers.length; ++i)
		{
			if ((mask & (1 << i)) != 0)
			{
				if (protocolHandlers[i] != null)
					throw new IllegalArgumentException(
						"Duplicate <handler> found, cannot proceed");
				protocolHandlers[i] = handler;
			}
		}
		capabilityMask |= mask;
	}

	/**
	 * Notifies all connection event handlers.
	 * 
	 * @param isConnected
	 *            Whether the connection has been established or it is about to
	 *            be closed
	 */
	private void raiseConnectionEvent(boolean isConnected)
	{
		connected = isConnected;
		Enumeration iter = connectionListeners.elements();
		while (iter.hasMoreElements())
		{
			if (isConnected)
				((ConnectionListener) iter.nextElement()).onConnection();
			else
				((ConnectionListener) iter.nextElement()).onDisconnection();
		}
	}

	/**
	 * Adds a connection listener, that will be notified of connection events.
	 * 
	 * @param listener
	 *            The connection event listener.
	 */
	public synchronized void addConnectionListener(ConnectionListener listener)
	{
		if (listener == null)
			throw new IllegalArgumentException(
				"Connection <listener> cannot be null");
		connectionListeners.addElement(listener);
	}

	/**
	 * Removes a previously added connection listener.
	 * 
	 * @param listener
	 *            The connection event listener.
	 */
	public synchronized void removeConnectionListener(
		ConnectionListener listener)
	{
		if (listener == null)
			throw new IllegalArgumentException(
				"Connection <listener> cannot be null");
		connectionListeners.removeElement(listener);
	}

	/**
	 * Initiates a connection using the connection provider specified during
	 * object construction.
	 * 
	 * @return The return value is true if the connection has been established,
	 *         false in case of errors.
	 */
	public synchronized boolean connect()
	{
		try
		{
			if (!provider.openConnection())
			{
				provider.closeConnection();
				return false;
			}
			reader
				.attachInput(
					provider.getInputStream(),
					(provider.getStreamProperties() & ConnectionProvider.AVAILABLE_IMPLEMENTED) != 0);
			writer.attachOutput(provider.getOutputStream());
		}
		catch (IOException e)
		{
			// invalidate reader, write and close connection
			terminate();
			return false;
		}
		// connection successful, go ahead
		connected = true;
		guard = false;
		if (SUPPORT_THREAD_DAEMON)
			readerThread.setDaemon(true);
		readerThread.start();
		return true;
	}

	/**
	 * Terminates the active connection and tries to release all the resources
	 * in use.
	 * 
	 * @param abort
	 *            Whether to drop connection immediately, or send a
	 *            disconnection chunk first
	 * 
	 * @return The return value is true for a clean disconnection, false if some
	 *         resources couldn't be released.
	 */
	public synchronized boolean disconnect(boolean abort)
	{
		if (!connected)
			return true;
		if (!abort)
			try
			{
				writer.sendDisconnect();
			}
			catch (IOException e)
			{
			}
		// signal thread to exit, wait and try to kill it if still alive
		reader.cancel();
		try
		{
			Thread.sleep(1000);
		}
		catch (Exception e)
		{
		}
		if (readerThread.isAlive())
		{
			if (SUPPORT_THREAD_INTERRUPT)
				readerThread.interrupt();
		}
		// do we have to cleanup?
		if (!guard)
		{
			guard = true;
			cancelHandlers();
			terminate();
			raiseConnectionEvent(false);
			// try again to kill thread, after closing the streams
			if (readerThread.isAlive())
			{
				try
				{
					Thread.sleep(1000);
					if (SUPPORT_THREAD_INTERRUPT)
						readerThread.interrupt();
				}
				catch (Exception e)
				{
				}
			}
			disposeHandlers();
		}
		return !readerThread.isAlive();
	}

	/**
	 * Perform tasks in order to terminate the connection.
	 */
	private void terminate()
	{
		reader.detachInput();
		writer.detachOutput();
		try
		{
			provider.closeConnection();
		}
		catch (IOException e)
		{
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/**
	 * Whether or not connected, at provider level
	 * 
	 * @return The return value is true if connected, false otherwise
	 */
	public boolean isConnected()
	{
		return connected;
	}

	/**
	 * Implements the loop that reads incoming chunks from the input stream and
	 * calls the appropriate handlers. This runs in a separate thread.
	 * 
	 * @see java.lang.Runnable#run()
	 */
	public void run()
	{
		raiseConnectionEvent(true);
		try
		{
			// go ahead and start receiving chunks and dispatching events
			while (reader.receiveChunk())
			{

			}
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
		drop();
	}

	/**
	 * Sends a packet to test the connection is still alive. A disconnection
	 * will be performed in case the packet could not be sent.
	 * 
	 * @param timeout
	 *            Time to wait for a response packet in milliseconds. Zero to
	 *            wait forever, a negative value to skip the wait.
	 * @return The return value is true if a response has arrived in time or no
	 *         wait was requested, false otherwise.
	 */
	public boolean test(long timeout)
	{
		if (!connected)
			return false;
		synchronized (tester)
		{
			tester.arrived = false;
			try
			{
				writer.sendPing();
			}
			catch (IOException e)
			{
				// TODO Auto-generated catch block
				e.printStackTrace();
				drop();
				return false;
			}
			if (timeout < 0)
				return true;
			try
			{
				tester.wait(timeout);
			}
			catch (InterruptedException e)
			{
				return false;
			}
			if (!tester.arrived)
				System.err.println("ping timed out");
			return tester.arrived;
		}
	}

	private void drop()
	{
		synchronized (this)
		{
			if (!guard)
			{
				guard = true;
				cancelHandlers();
				terminate();
				raiseConnectionEvent(false);
				disposeHandlers();
			}
		}
	}

	/**
	 * Sends packet to configure the protocol capabilities
	 * 
	 * @param timeout
	 *            Time to wait for a response packet in milliseconds. Zero to
	 *            wait forever, a negative value to skip the wait.
	 * @return The return value is true if a response has arrived in time or no
	 *         wait was requested, false otherwise.
	 */
	public boolean setup(long timeout)
	{
		if (!connected)
			return false;
		synchronized (setupper)
		{
			setupper.arrived = false;
			try
			{
				// send configuration
				writer.sendSetup(capabilityMask, maxSpeed);
			}
			catch (IOException e)
			{
				// TODO Auto-generated catch block
				e.printStackTrace();
				drop();
				return false;
			}
			if (timeout < 0)
				return true;
			try
			{
				setupper.wait(timeout);
			}
			catch (InterruptedException e)
			{
				return false;
			}
			if (!setupper.arrived)
				System.err.println("setup timed out");
			return setupper.arrived;
		}
	}

	// Listener interface

	public void onChunk(int type)
	{
		// TODO resetta il timeout di inattivit della connessione
		// DEBUG System.out.println("rx chunk " + type);
	}

	public void onPong()
	{
		// do nothing
		System.out.println("rx Pong");

		synchronized (tester)
		{
			tester.arrived = true;
			tester.notify();
		}
	}

	public void onSetup(int capabilityMask, int maxSpeed)
	{
		// DEBUG System.out.println("rx Setup " + capabilityMask + "," + maxSpeed);
		synchronized (this)
		{
			this.maxSpeed = maxSpeed;
			this.capabilityMask &= capabilityMask;
			// cancel handlers for unsupported capabilities
			for (int i = 0; i < protocolHandlers.length; ++i)
			{
				if ((this.capabilityMask & (1 << i)) == 0)
				{
					if (protocolHandlers[i] != null)
						protocolHandlers[i].cancelHandler();
				}
			}
			// signal setup arrived
			synchronized (setupper)
			{
				setupper.arrived = true;
				setupper.notify();
			}
		}
	}

	public void onPing()
	{
		// reply to ping
		try
		{
			writer.sendPong();
		}
		catch (IOException e)
		{
			reader.cancel();
		}
	}

	public void onDisconnect()
	{
		// will drop connection
	}
}
