package communication.handler.bus;

/**
 * Supports the creation of messages to be transferred on the system bus.
 * 
 * @author Paolo
 */
public class BusData
{
	/**
	 * Converts an internal module identifier to an I2C slave address
	 * 
	 * @param module
	 *            The module indentifier
	 * @return The bus address
	 * 
	 * @see communication.handler.internal.InternalModule
	 */
	public static byte internalAddress(int module)
	{
		if (module < 0 || module > 7)
			throw new IllegalArgumentException("invalid module identifier");
		return (byte) (module + 40);
	}

	int idx;
	int num;
	int[] slave;
	boolean[] flag;
	byte[][] mesg;
	byte[][] result;
	boolean cancelled;
	boolean reply;

	void checkAdd()
	{
		if (idx >= num)
			throw new IllegalArgumentException("message number exceeded");
	}

	void checkSlave(int addr)
	{
		if (addr > 127)
			throw new IllegalArgumentException("invalid slave addresses");
	}

	void checkRead(int index)
	{
		if (!flag[index])
			throw new IllegalArgumentException("not a read message");
	}

	void checkWrite(int index)
	{
		if (flag[index])
			throw new IllegalArgumentException("not a write message");
	}

	void checkValid()
	{
		if (idx < num)
			throw new IllegalArgumentException("incomplete message data");
	}

	void checkReply()
	{
		if (!reply)
			throw new IllegalArgumentException("must wait for reply first");
	}

	/**
	 * Creates a sequence of messages with the specified length.
	 * 
	 * @param numMessages
	 *            The number of messages in the sequence
	 */
	public BusData(int numMessages)
	{
		idx = 0;
		num = numMessages;
		slave = new int[num];
		flag = new boolean[num];
		mesg = new byte[num][];
		result = null;
		cancelled = false;
		reply = false;
	}

	/**
	 * Marks this message sequence to be reused for transfers.
	 */
	public void recycle()
	{
		checkValid();
		result = null;
		cancelled = false;
		reply = false;
		for (int i = 0; i < num; ++i)
		{
			if (flag[i])
			{
				int length = mesg[i].length;
				mesg[i] = new byte[]
				{
					(byte) length
				};
			}
		}
	}

	/**
	 * Appends a write message to the sequence.
	 * 
	 * @param addr
	 *            The address of an I2C slave device
	 * @param data
	 *            The buffer holding data to be written
	 */
	public void addWriteMessage(int addr, byte[] data)
	{
		checkAdd();
		checkSlave(addr);
		slave[idx] = addr;
		flag[idx] = false;
		mesg[idx] = data;
		++idx;
	}

	/**
	 * Appends a read message to the sequence.
	 * 
	 * @param addr
	 *            The address of an I2C slave device
	 * @param length
	 *            The length of data to be read
	 * @return The index of this message to retrieve the actual data after the
	 *         transfer completes
	 * 
	 * @see #getReadData(int)
	 */
	public int addReadMessage(int addr, int length)
	{
		checkAdd();
		checkSlave(addr);
		slave[idx] = addr;
		flag[idx] = true;
		mesg[idx] = new byte[]
		{
			(byte) length
		};
		return idx++;
	}

	/**
	 * Retrieves data transferred in a read message.
	 * 
	 * @param index
	 *            The index of a read message in this sequence
	 * @return The buffer holding tha actual data read
	 */
	public byte[] getReadData(int index)
	{
		checkReply();
		checkRead(index);
		return mesg[index];
	}

	/**
	 * Retrieves data transferred in a write message.
	 * 
	 * @param index
	 *            The index of a write message in this sequence
	 * @return The buffer holding tha actual data read
	 */
	public byte[] getWriteData(int index)
	{
		checkWrite(index);
		return mesg[index];
	}

	/**
	 * Waits until the sequence has been actually transferred.
	 * 
	 * @return The return value is true if the transfer completed successfully,
	 *         false in case of errors
	 * 
	 * @throws InterruptedException
	 *             If communication is interrupted before completion
	 */
	public synchronized boolean waitForResult() throws InterruptedException
	{
		if (cancelled)
			throw new InterruptedException("Operation cancelled");

		// wait for result and evaluate success of the operation
		if (!reply)
			wait();
		if (cancelled)
			throw new InterruptedException("Operation cancelled");

		if (result == null)
			return false;
		// success
		int j = 0;
		for (int i = 0; i < num; ++i)
		{
			if (j > result.length)
				throw new IllegalArgumentException(
					"less messages than expected in reply");

			if (flag[i])
			{
				if (((int) mesg[i][0] & 255) != result[j].length)
					throw new IllegalArgumentException(
						"message data length mismatch");
				mesg[i] = result[j];
				++j;
			}
		}
		if (j < result.length)
			throw new IllegalArgumentException(
				"more messages than expected in reply");
		return true;
	}
}
