package communication.handler.bus;

import java.io.IOException;
import java.util.Vector;

import communication.handler.AbstractCapabilityHandler;
import communication.protocol.BusListener;
import communication.protocol.Capability;
import communication.protocol.ChunkReader;
import communication.protocol.ChunkWriter;

/**
 * Handler for the Bus protocol capability, that provides generic access to the
 * I2C bus.
 * <p>
 * Public methods for users of this class are:
 * <ul>
 * <li>{@link #busTransfer}
 * <li>{@link #busPause}
 * </ul>
 * 
 * @author Paolo
 */
public class BusHandler extends AbstractCapabilityHandler implements
	BusListener
{
	private Vector queue;

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.handler.CapabilityHandler#initializeHandler(communication.protocol.ChunkReader,
	 *      communication.protocol.ChunkWriter)
	 */
	public int initializeHandler(ChunkReader reader, ChunkWriter writer)
	{
		super.initializeHandler(reader, writer);
		queue = new Vector(10);
		reader.setBusListener(this);
		return Capability.BUS;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.handler.CapabilityHandler#disposeHandler()
	 */
	public void disposeHandler()
	{
		reader.setBusListener(null);
		queue = null;
		super.disposeHandler();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.handler.CapabilityHandler#cancelHandler()
	 */
	public void cancelHandler()
	{
		super.cancelHandler();
		// notify any blocking method, return with exception
		while (queue.size() > 0)
		{
			BusData busData = (BusData) queue.firstElement();
			queue.removeElementAt(0);
			synchronized (busData)
			{
				busData.reply = true;
				busData.result = null;
				busData.cancelled = true;
				busData.notify();
			}
		}
	}

	public void onBusResult(byte[][] data)
	{
		BusData busData = (BusData) queue.firstElement();
		queue.removeElementAt(0);
		synchronized (busData)
		{
			busData.reply = true;
			busData.result = data;
			busData.notify();
		}
	}

	/**
	 * Queues a set of I2C messages to be transferred in a single transaction on
	 * the system bus.
	 * 
	 * @param data
	 *            The object containing the messages to be transferred on the
	 *            bus
	 * 
	 * @throws IOException
	 *             In case the transport layer fails
	 * @throws InterruptedException
	 *             If communication is interrupted before completion
	 * @see BusData
	 */
	public void busTransfer(BusData data) throws IOException,
		InterruptedException
	{
		synchronized (data)
		{
			data.result = null;
			data.reply = true;
			data.cancelled = true;
			data.checkValid();
			if (cancelled)
				throw new InterruptedException("Operation cancelled");
			// ok, proceed
			data.reply = false;
			data.cancelled = false;
			queue.addElement(data);
			writer.sendBusTransfer(data.slave, data.flag, data.mesg);
		}
	}

	/**
	 * Queues a pause between bus transfers, for timing purposes. This should be
	 * intended as a minimum value.
	 * 
	 * @param millisecs
	 *            The pause duration in milliseconds
	 * 
	 * @throws IOException
	 *             In case the transport layer fails
	 * @throws InterruptedException
	 *             If communication is interrupted before completion
	 */
	public void busPause(int millisecs) throws IOException,
		InterruptedException
	{
		if (cancelled)
			throw new InterruptedException("Operation cancelled");
		writer.sendBusPause(millisecs);
	}
}
