package communication.handler.internal;

import java.io.IOException;

import communication.handler.AbstractCapabilityHandler;
import communication.protocol.Capability;
import communication.protocol.ChunkReader;
import communication.protocol.ChunkWriter;
import communication.protocol.InternalListener;

/**
 * Handler for the Internal protocol capability, that provides access to the
 * robot's internal modules over the I2C bus.
 * <p>
 * Public methods for users of this class are:
 * <ul>
 * <li>{@link #readRegister}
 * <li>{@link #writeRegister}
 * </ul>
 * Both are internally synchronized, independently of each other.
 * 
 * @author Paolo
 */
public class InternalHandler extends AbstractCapabilityHandler implements
	InternalListener
{
	private Object readSync;

	private Object writeSync;

	private class WriteRegEvent
	{
		public boolean result;
	}

	private WriteRegEvent writeEvent;

	private class ReadRegEvent
	{
		public int result[];
	}

	private ReadRegEvent readEvent;

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.handler.CapabilityHandler#initializeHandler(communication.protocol.ChunkReader,
	 *      communication.protocol.ChunkWriter)
	 */
	public int initializeHandler(ChunkReader reader, ChunkWriter writer)
	{
		super.initializeHandler(reader, writer);
		readSync = new Object();
		writeSync = new Object();
		writeEvent = new WriteRegEvent();
		readEvent = new ReadRegEvent();
		reader.setInternalListener(this);
		return Capability.INTERNAL;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.handler.CapabilityHandler#disposeHandler()
	 */
	public void disposeHandler()
	{
		reader.setInternalListener(null);
		writeEvent = null;
		readEvent = null;
		super.disposeHandler();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.handler.CapabilityHandler#cancelHandler()
	 */
	public void cancelHandler()
	{
		super.cancelHandler();
		// notify any blocking method, return with exception
		onInternalReadError();
		onInternalWriteError();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.protocol.InternalListener#onInternalWriteError()
	 */
	public void onInternalWriteError()
	{
		synchronized (writeEvent)
		{
			writeEvent.result = false;
			writeEvent.notify();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.protocol.InternalListener#onInternalWriteOk()
	 */
	public void onInternalWriteOk()
	{
		synchronized (writeEvent)
		{
			writeEvent.result = true;
			writeEvent.notify();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.protocol.InternalListener#onInternalReadError()
	 */
	public void onInternalReadError()
	{
		synchronized (readEvent)
		{
			readEvent.result = null;
			readEvent.notify();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.protocol.InternalListener#onInternalRegData1(int)
	 */
	public void onInternalRegData1(int value)
	{
		synchronized (readEvent)
		{
			readEvent.result[0] = value;
			readEvent.notify();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.protocol.InternalListener#onInternalRegData2(int, int)
	 */
	public void onInternalRegData2(int value1, int value2)
	{
		synchronized (readEvent)
		{
			readEvent.result[0] = value1;
			readEvent.result[1] = value2;
			readEvent.notify();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.protocol.InternalListener#onInternalRegData3(int, int,
	 *      int)
	 */
	public void onInternalRegData3(int value1, int value2, int value3)
	{
		synchronized (readEvent)
		{
			readEvent.result[0] = value1;
			readEvent.result[1] = value2;
			readEvent.result[2] = value3;
			readEvent.notify();
		}
	}

	/**
	 * Performs a write operation to an internal module registers. Caller is
	 * blocked until completion. Behavior is undefined in case of exceptions.
	 * 
	 * @param module
	 *            An integer identifier of the module (0-7)
	 * @param register
	 *            The index of a module register (0-31)
	 * @param value
	 *            Array of integer values (0-255) to write. Valid lengths are 1,
	 *            2, 3.
	 * @return The return value is true when successful, false if module did not
	 *         respond.
	 * 
	 * @throws IOException
	 *             In case the transport layer fails
	 * @throws InterruptedException
	 *             If communication is interrupted before completion
	 * @see InternalModule
	 */
	public boolean writeRegister(int module, int register, int value[])
		throws IOException, InterruptedException
	{
		if (cancelled)
			throw new InterruptedException("Operation cancelled");
		synchronized (writeSync)
		{
			boolean ret;
			synchronized (writeEvent)
			{
				// send appropriate write command
				switch (value.length)
				{
				case 1:
					writer.sendWriteInternalReg1(module, register, value[0]);
					break;
				case 2:
					writer.sendWriteInternalReg2(module, register, value[0],
						value[1]);
					break;
				case 3:
					writer.sendWriteInternalReg3(module, register, value[0],
						value[1], value[2]);
					break;
				default:
					throw new IllegalArgumentException(
						"Unsupported array length for argument <value>");
				}
				writeEvent.wait();
				ret = writeEvent.result;
			}
			if (cancelled)
				throw new InterruptedException("Operation cancelled");
			return ret;
		}
	}

	/**
	 * Performs a read operation from an internal module registers. Caller is
	 * blocked until completion. Behavior is undefined in case of exceptions.
	 * 
	 * @param module
	 *            An integer identifier of the module (0-7)
	 * @param register
	 *            The index of a module register (0-31)
	 * @param value
	 *            Array to be filled with integer values (0-255). Valid lengths
	 *            are 1, 2, 3.
	 * @return The return value is true when successful, false if module did not
	 *         respond.
	 * 
	 * @throws IOException
	 *             In case the transport layer fails
	 * @throws InterruptedException
	 *             If communication is interrupted before completion
	 * @see InternalModule
	 */
	public boolean readRegister(int module, int register, int value[])
		throws IOException, InterruptedException
	{
		if (cancelled)
			throw new InterruptedException("Operation cancelled");
		synchronized (readSync)
		{
			boolean ret;
			synchronized (readEvent)
			{
				// prepare result and send appropriate read command
				switch (value.length)
				{
				case 1:
					readEvent.result = value;
					writer.sendReadInternalReg1(module, register);
					break;
				case 2:
					readEvent.result = value;
					writer.sendReadInternalReg2(module, register);
					break;
				case 3:
					readEvent.result = value;
					writer.sendReadInternalReg3(module, register);
					break;
				default:
					throw new IllegalArgumentException(
						"Unsupported array length for argument <value>");
				}
				// wait for result and evaluate success of the operation
				readEvent.wait();
				ret = (readEvent.result != null);
				// release internal reference to data
				readEvent.result = null;
			}
			if (cancelled)
				throw new InterruptedException("Operation cancelled");
			return ret;
		}
	}
}
