package communication.handler.transfer;

import java.io.IOException;

/**
 * Interface to the remote directory enumeration protocol.
 * 
 * @author Paolo
 */
public interface DirectoryEnumeration
{
	/**
	 * Constant for unknown or unsupported directory entries.
	 */
	public static int UNKNOWN = 0;
	/**
	 * Constant for regular file entries.
	 */
	public static int FILE = 1;
	/**
	 * Constant for directory type entries.
	 */
	public static int DIRECTORY = 2;
	/**
	 * Constant for symbolic link entries.
	 */
	public static int LINK = 3;
	/**
	 * Constant for character device entries.
	 */
	public static int CHAR_DEVICE = 4;
	/**
	 * Constant for block device entries.
	 */
	public static int BLOCK_DEVICE = 5;
	/**
	 * Constant for FIFO type entries.
	 */
	public static int FIFO = 6;
	/**
	 * Constant for socket type entries.
	 */
	public static int SOCKET = 7;

	/**
	 * Looks for the next entry in the open directory.
	 * 
	 * @return The return value is true if a new entry has been found or false
	 *         otherwise.
	 * @throws IOException
	 *             In case the transport layer fails
	 * @throws InterruptedException
	 *             If communication is interrupted before completion
	 */
	public boolean nextEntry() throws IOException, InterruptedException;

	/**
	 * Retrieves the current entry in a directory.
	 * 
	 * @return The name of the current entry (relative path).
	 */
	public String getEntry();

	/**
	 * Retrieves the entry type for the current entry (whether it is a regular
	 * file, a sub-directory or a special file).
	 * 
	 * @return The type of the current entry.
	 * 
	 * @see #UNKNOWN
	 * @see #FILE
	 * @see #DIRECTORY
	 * @see #LINK
	 * @see #CHAR_DEVICE
	 * @see #BLOCK_DEVICE
	 * @see #FIFO
	 * @see #SOCKET
	 */
	public int getEntryType();

	/**
	 * Terminates the enumeration process. This must be called at the end of the
	 * enumeration to free remote resources.
	 * 
	 * @throws IOException
	 *             In case the transport layer fails
	 * @throws InterruptedException
	 *             If communication is interrupted before completion
	 */
	public void close() throws IOException, InterruptedException;
}
