package communication.handler.transfer;

import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.io.OutputStream;

import communication.handler.AbstractCapabilityHandler;
import communication.protocol.Capability;
import communication.protocol.ChunkReader;
import communication.protocol.ChunkWriter;
import communication.protocol.TransferListener;

/**
 * Handler for the Transfer protocol capability, that provides remote access to
 * the robot's file system.
 * <p>
 * Public methods for users of this class are:
 * <ul>
 * <li>{@link #openInputStream}
 * <li>{@link #openOutputStream}
 * </ul>
 * They provide usual stream abstractions for remote files.
 * 
 * @author Paolo
 */
public class TransferHandler extends AbstractCapabilityHandler implements
	TransferListener
{
	private class OpenEvent
	{
		public int token;
		public int result;
	}

	private class FileEvent
	{
		public boolean assigned;
		public int result;
		public byte[] data;
		public String entry;
		public int type;
		public FileEvent async; // non-null if async file
	}

	private OpenEvent openEvent;
	private FileEvent[] fileEvents;
	private static final int MAX_TOKENS = 10;

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.handler.CapabilityHandler#initializeHandler(communication.protocol.ChunkReader,
	 *      communication.protocol.ChunkWriter)
	 */
	public int initializeHandler(ChunkReader reader, ChunkWriter writer)
	{
		super.initializeHandler(reader, writer);
		openEvent = new OpenEvent();
		fileEvents = new FileEvent[MAX_TOKENS];
		for (int i = 0; i < MAX_TOKENS; ++i)
			fileEvents[i] = new FileEvent();
		reader.setTransferListener(this);
		return Capability.TRANSFER;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.handler.CapabilityHandler#disposeHandler()
	 */
	public void disposeHandler()
	{
		reader.setTransferListener(null);
		openEvent = null;
		fileEvents = null;
		super.disposeHandler();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.handler.CapabilityHandler#cancelHandler()
	 */
	public void cancelHandler()
	{
		super.cancelHandler();
		// notify any blocking method, return with exception
		synchronized (openEvent)
		{
			openEvent.result = RES_ERR_EXCEPTION;
			openEvent.notify();
		}
		for (int i = 0; i < MAX_TOKENS; ++i)
			synchronized (fileEvents[i])
			{
				fileEvents[i].result = RES_ERR_EXCEPTION;
				fileEvents[i].data = null;
				fileEvents[i].entry = null;
				fileEvents[i].notifyAll();

				if (fileEvents[i].async != null)
				{
					synchronized (fileEvents[i].async)
					{
						fileEvents[i].async.result = RES_ERR_EXCEPTION;
						fileEvents[i].async.data = null;
						fileEvents[i].async.notifyAll();
					}
				}
			}
	}

	// Protocol implementation

	public void onTransferResult(int token, int code)
	{
		// remap tokens internally (0 to -1, 1... to 0...)
		--token;
		if (token < 0 || !fileEvents[token].assigned)
		{
			synchronized (openEvent)
			{
				openEvent.token = token;
				openEvent.result = code;
				openEvent.notify();
			}
		}
		else
		{
			if (code >= RES_ASYNC_MIN && code <= RES_ASYNC_MAX)
			{
				synchronized (fileEvents[token].async)
				{
					if (fileEvents[token].async.result != RES_PENDING)
						try
						{
							fileEvents[token].async.wait();
						}
						catch (InterruptedException e)
						{
							e.printStackTrace();
						}
					fileEvents[token].async.result = code;
					fileEvents[token].async.data = null;
					fileEvents[token].async.notify();
				}
			}
			else
			{
				synchronized (fileEvents[token])
				{
					fileEvents[token].result = code;
					fileEvents[token].data = null;
					fileEvents[token].entry = null;
					fileEvents[token].notify();
				}
			}
		}
	}

	public void onTransferData(int token, byte[] data)
	{
		--token; // TODO handle invalid token!
		if (fileEvents[token].async != null)
		{
			synchronized (fileEvents[token].async)
			{
				if (fileEvents[token].async.result != RES_PENDING)
					try
					{
						fileEvents[token].async.wait();
					}
					catch (InterruptedException e)
					{
						e.printStackTrace();
					}
				fileEvents[token].async.result = RES_OK;
				fileEvents[token].async.data = data;
				fileEvents[token].async.notify();
			}
		}
		else
		{
			synchronized (fileEvents[token])
			{
				fileEvents[token].result = RES_OK;
				fileEvents[token].data = data;
				fileEvents[token].notify();
			}
		}
	}

	public void onDirectoryEntry(int token, String name, int type)
	{
		--token; // TODO handle invalid token!
		synchronized (fileEvents[token])
		{
			fileEvents[token].result = RES_OK;
			fileEvents[token].entry = name;
			fileEvents[token].type = type;
			fileEvents[token].notify();
		}
	}

	public void onTransferOpen(int flags, String name)
	{
		// TODO server interface
	}

	public void onTransferClose(int token)
	{
		// TODO server interface
	}

	public void onTransferWrite(int token, byte[] data)
	{
		// TODO server interface
	}

	public void onTransferRead(int token, int size)
	{
		// TODO server interface
	}

	public void onDirectoryNext()
	{
		// TODO server interface
	}

	// Public interface

	private static final int FLAG_READ = 1;
	private static final int FLAG_WRITE = 2;
	private static final int FLAG_ENUM = 16;
	private static final int FLAG_PROC = 32;
	private static final int FLAG_ASYNC = 64;

	static final int RES_OK = 0;
	static final int RES_EOF = 255;
	static final int RES_PENDING = 254;
	static final int RES_ERR_READ = 1;
	static final int RES_ERR_WRITE = 2;
	static final int RES_ERR_OPEN = 3;
	static final int RES_ERR_CLOSE = 4;
	static final int RES_ERR_TOKEN = 5;
	static final int RES_ERR_TYPE = 6;
	static final int RES_ERR_FLUSH = 7;
	// (64 < result < 128) reserved for async replies
	private static final int RES_ASYNC_MIN = 64;
	static final int RES_ASYNC_READ_ERR = 65;
	static final int RES_ASYNC_READ_EOF = 66;
	private static final int RES_ASYNC_MAX = 127;
	// internal result code to signal cancellation
	private static final int RES_ERR_EXCEPTION = -1;

	/**
	 * Enumerates the entries of a remote directory.
	 * 
	 * @param vpath
	 *            The directory path on the remote virtual file system (relative
	 *            paths point to the RW flash partition).
	 * @return An enumeration object if successful, null in case of errors.
	 * 
	 * @throws IOException
	 *             In case the transport layer fails
	 * @throws InterruptedException
	 *             If communication is interrupted before completion
	 */
	public synchronized DirectoryEnumeration listDirectory(String vpath)
		throws IOException, InterruptedException
	{
		if (cancelled)
			throw new InterruptedException("Operation cancelled");
		synchronized (openEvent)
		{
			writer.sendFileOpen(FLAG_ENUM, vpath);
			openEvent.wait();
			if (cancelled)
				throw new InterruptedException("Operation cancelled");
			if (openEvent.result == RES_OK)
				return new DirEnum(openEvent.token);
		}
		return null;
	}

	private class DirEnum implements DirectoryEnumeration
	{
		private int token;
		private FileEvent event;
		private String entry;
		private int type;

		public DirEnum(int token)
		{
			this.token = token + 1;
			event = fileEvents[token];
			event.assigned = true;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see communication.handler.transfer.DirectoryEnumeration#hasMoreEntries()
		 */
		public boolean nextEntry() throws IOException, InterruptedException
		{
			synchronized (event)
			{
				writer.sendDirectoryNext(token);
				event.wait();
				if (event.result == RES_OK)
				{
					// ok, we got data, copy it back to the caller
					entry = event.entry;
					type = event.type;
					return true;
				}
				else if (event.result == RES_ERR_EXCEPTION)
					throw new InterruptedException("Operation cancelled");
				else if (event.result != RES_EOF)
					throw new IOException("Transfer error " + event.result);
			}
			return false;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see communication.handler.transfer.DirectoryEnumeration#getEntry()
		 */
		public String getEntry()
		{
			return entry;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see communication.handler.transfer.DirectoryEnumeration#getEntryType()
		 */
		public int getEntryType()
		{
			return type;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see communication.handler.transfer.DirectoryEnumeration#close()
		 */
		public void close() throws IOException, InterruptedException
		{
			synchronized (event)
			{
				writer.sendFileClose(token);
				event.wait();
				event.assigned = false;
				if (event.result == RES_ERR_EXCEPTION)
					throw new InterruptedException("Operation cancelled");
				else if (event.result != RES_OK)
					throw new IOException("Transfer error " + event.result);
			}
		}
	}

	/**
	 * Creates an empty remote file for writing.
	 * 
	 * @param vpath
	 *            The file path on the remote virtual file system (relative
	 *            paths point to the RW flash partition).
	 * @return The opened output stream if successful, null in case of errors.
	 * 
	 * @throws IOException
	 *             In case the transport layer fails
	 * @throws InterruptedException
	 *             If communication is interrupted before completion
	 */
	public synchronized FileOutputStream openOutputStream(String vpath)
		throws IOException, InterruptedException
	{
		if (cancelled)
			throw new InterruptedException("Operation cancelled");
		synchronized (openEvent)
		{
			writer.sendFileOpen(FLAG_WRITE, vpath);
			openEvent.wait();
			if (cancelled)
				throw new InterruptedException("Operation cancelled");
			if (openEvent.result == RES_OK)
				return new FileOutputStream(openEvent.token);
		}
		return null;
	}

	/**
	 * Executes a remote command and opens its input stream for writing.
	 * 
	 * @param command
	 *            The command to execute on the remote system.
	 * @return The opened stream if successful, null in case of errors.
	 * 
	 * @throws IOException
	 *             In case the transport layer fails
	 * @throws InterruptedException
	 *             If communication is interrupted before completion
	 */
	public synchronized OutputStream openCommandInput(String command)
		throws IOException, InterruptedException
	{
		if (cancelled)
			throw new InterruptedException("Operation cancelled");
		synchronized (openEvent)
		{
			writer.sendFileOpen(FLAG_WRITE | FLAG_PROC, command);
			openEvent.wait();
			if (cancelled)
				throw new InterruptedException("Operation cancelled");
			if (openEvent.result == RES_OK)
				return new FileOutputStream(openEvent.token);
		}
		return null;
	}

	private class FileOutputStream extends OutputStream
	{
		private int token;
		private FileEvent event;
		private byte[] buf = new byte[1];

		public FileOutputStream(int token)
		{
			this.token = token + 1;
			event = fileEvents[token];
			event.assigned = true;
		}

		public void write(int b) throws IOException
		{
			buf[0] = (byte) (b & 0xFF);
			write(buf);
		}

		public void write(byte[] b, int off, int len) throws IOException
		{
			byte[] buffer = new byte[len];
			if (len > 0)
				System.arraycopy(b, off, buffer, 0, len);
			write(buffer);
		}

		public void close() throws IOException
		{
			synchronized (event)
			{
				if (token < 0)
					throw new EOFException("stream has been closed");
				writer.sendFileClose(token);
				token = -1;
				try
				{
					event.wait();
				}
				catch (InterruptedException e)
				{
					event.assigned = false;
					e.printStackTrace();
					throw new InterruptedIOException(e.toString());
				}
				event.assigned = false;
				if (event.result != RES_OK)
					throw new IOException("Transfer error " + event.result);
			}
		}

		public void write(byte[] b) throws IOException
		{
			synchronized (event)
			{
				if (token < 0)
					throw new EOFException("stream has been closed");
				if (b.length == 0)
					return;
				writer.sendFileWrite(token, b);
				try
				{
					event.wait();
				}
				catch (InterruptedException e)
				{
					e.printStackTrace();
					throw new InterruptedIOException(e.toString());
				}
				if (event.result != RES_OK)
					throw new IOException("Transfer error " + event.result);
			}
		}

		public void flush() throws IOException
		{
			synchronized (event)
			{
				if (token < 0)
					throw new EOFException("stream has been closed");
				writer.sendFileFlush(token);
				try
				{
					event.wait();
				}
				catch (InterruptedException e)
				{
					e.printStackTrace();
					throw new InterruptedIOException(e.toString());
				}
				if (event.result != RES_OK)
					throw new IOException("Transfer error " + event.result);
			}
		}
	}

	/**
	 * Opens a remote file for reading, with optional asynchronous behavior.
	 * Asynchronous operations do not block all the threads involved in the
	 * communication, but only the thread accessing the stream.
	 * 
	 * @param vpath
	 *            The file path on the remote virtual file system (relative
	 *            paths point to the RW flash partition).
	 * @param async
	 *            A flag that specifies whether to use asynchronous read
	 *            operations
	 * @return The opened input stream if successful, null in case of errors.
	 * 
	 * @throws IOException
	 *             In case the transport layer fails
	 * @throws InterruptedException
	 *             If communication is interrupted before completion
	 */
	public synchronized InputStream openInputStream(String vpath, boolean async)
		throws IOException, InterruptedException
	{
		return openForInput(vpath, async, false);
	}

	/**
	 * Opens a remote file for reading, with synchrounous behavior.
	 * 
	 * @param vpath
	 *            The file path on the remote virtual file system (relative
	 *            paths point to the RW flash partition).
	 * @return The opened input stream if successful, null in case of errors.
	 * 
	 * @throws IOException
	 *             In case the transport layer fails
	 * @throws InterruptedException
	 *             If communication is interrupted before completion
	 */
	public synchronized InputStream openInputStream(String vpath)
		throws IOException, InterruptedException
	{
		return openForInput(vpath, false, false);
	}

	/**
	 * Executes a remote command and opens its output stream for reading, with
	 * optional asynchronous behavior. Asynchronous operations do not block all
	 * the threads involved in the communication, but only the thread accessing
	 * the stream.
	 * 
	 * @param command
	 *            The command to execute on the remote system.
	 * @param async
	 *            A flag that specifies whether to use asynchronous read
	 *            operations
	 * @return The opened stream if successful, null in case of errors.
	 * 
	 * @throws IOException
	 *             In case the transport layer fails
	 * @throws InterruptedException
	 *             If communication is interrupted before completion
	 */
	public synchronized InputStream openCommandOutput(String command,
		boolean async) throws IOException, InterruptedException
	{
		return openForInput(command, async, true);
	}

	/**
	 * Executes a remote command and opens its output stream for reading, with
	 * synchronous behavior.
	 * 
	 * @param command
	 *            The command to execute on the remote system.
	 * @return The opened stream if successful, null in case of errors.
	 * 
	 * @throws IOException
	 *             In case the transport layer fails
	 * @throws InterruptedException
	 *             If communication is interrupted before completion
	 */
	public synchronized InputStream openCommandOutput(String command)
		throws IOException, InterruptedException
	{
		return openForInput(command, false, true);
	}

	protected FileInputStream openForInput(String path, boolean async,
		boolean process) throws IOException, InterruptedException
	{
		if (cancelled)
			throw new InterruptedException("Operation cancelled");
		int flags = FLAG_READ;
		if (async)
			flags |= FLAG_ASYNC;
		if (process)
			flags |= FLAG_PROC;
		synchronized (openEvent)
		{
			writer.sendFileOpen(flags, path);
			openEvent.wait();
			if (cancelled)
				throw new InterruptedException("Operation cancelled");
			if (openEvent.result == RES_OK)
				return new FileInputStream(openEvent.token, async);
		}
		return null;
	}

	private class FileInputStream extends InputStream
	{
		private int token;
		private FileEvent event;
		private byte[] buf = new byte[1];

		public FileInputStream(int token, boolean async)
		{
			this.token = token + 1;
			event = fileEvents[token];
			event.assigned = true;
			if (async)
				event.async = new FileEvent();
			else
				event.async = null;
		}

		public int read() throws IOException
		{
			int ret = read(buf, 0, 1);
			if (ret <= 0)
				return -1;
			return (int) buf[0] & 0xFF;
		}

		public void close() throws IOException
		{
			synchronized (event)
			{
				if (token < 0)
					throw new EOFException("stream has been closed");
				writer.sendFileClose(token);
				token = -1;
				try
				{
					event.wait();
				}
				catch (InterruptedException e)
				{
					event.assigned = false;
					e.printStackTrace();
					throw new InterruptedIOException(e.toString());
				}
				event.assigned = false;
				if (event.result != RES_OK)
					throw new IOException("Transfer error " + event.result);
			}
		}

		public int read(byte[] b, int off, int len) throws IOException
		{
			int size = -1;
			boolean pending = false;
			synchronized (event)
			{
				if (token < 0)
					throw new EOFException("stream has been closed");
				if (len <= 0)
					return 0;
				writer.sendFileRead(token, len);
				try
				{
					event.wait();
				}
				catch (InterruptedException e)
				{
					e.printStackTrace();
					throw new InterruptedIOException(e.toString());
				}
				if (event.result == RES_PENDING)
				{
					pending = true; // postponed result
				}
				else if (event.result == RES_OK)
				{
					// ok, we got data, copy it back to the caller
					size = event.data.length;
					if (size > 0)
						System.arraycopy(event.data, 0, b, off, size);
				}
				else if (event.result != RES_EOF)
					throw new IOException("Transfer error " + event.result);
			}
			// with async result, we must release event mutex and use another
			if (pending)
			{
				synchronized (event.async)
				{
					// sync with receiver
					event.async.result = RES_PENDING;
					event.async.notify();
					// check reply
					if (event.async.result == RES_PENDING)
						try
						{
							event.async.wait();
						}
						catch (InterruptedException e)
						{
							e.printStackTrace();
							throw new InterruptedIOException(e.toString());
						}
					if (event.async.result == RES_OK)
					{
						// ok, we got data, copy it back to the caller
						size = event.async.data.length;
						if (size > 0)
							System.arraycopy(event.async.data, 0, b, off, size);
					}
					else if (event.async.result != RES_ASYNC_READ_EOF)
						throw new IOException("Transfer error "
							+ event.async.result);
				}
			}
			return size;
		}
	}
}
