package communication.handler.video;

import java.io.IOException;
import java.util.Stack;
import java.util.Vector;

import communication.handler.AbstractCapabilityHandler;
import communication.protocol.Capability;
import communication.protocol.ChunkReader;
import communication.protocol.ChunkWriter;
import communication.protocol.VideoListener;

/**
 * Handler for the Video protocol capability, that provides streaming of the
 * video images acquired by the robot.
 * <p>
 * Public methods for users of this class are:
 * <ul>
 * <li>{@link #configureVideo}
 * <li>{@link #startVideo}
 * <li>{@link #stopVideo}
 * <li>{@link #acquireBuffer}
 * <li>{@link #releaseBuffer}
 * </ul>
 * 
 * @author Paolo
 */
public class VideoHandler extends AbstractCapabilityHandler implements
	VideoListener
{
	private int bufferLength;

	private int frameWidth;

	private int frameHeight;

	private int pixelFormat;

	// private int frameRate;

	private int poolSize;

	private class State
	{
		public static final int IDLE = 0; // not configured

		public static final int READY = 1; // just configured

		public static final int ACTIVE = 2; // started

		public static final int INACTIVE = 3; // stopped

		public int id = IDLE;
	}

	private State state;

	private Stack bufferPool; // a LIFO should optimize cache usage

	private Vector bufferQueue; // buffers ready with video data

	private Object synchPool = new Object();
	private Object synchQueue = new Object();

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.handler.AbstractCapabilityHandler#initializeHandler(communication.protocol.ChunkReader,
	 *      communication.protocol.ChunkWriter)
	 */
	public int initializeHandler(ChunkReader reader, ChunkWriter writer)
	{
		state = new State();
		bufferPool = new Stack();
		bufferQueue = new Vector();
		super.initializeHandler(reader, writer);
		reader.setVideoListener(this);
		return Capability.VIDEO;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.handler.AbstractCapabilityHandler#disposeHandler()
	 */
	public void disposeHandler()
	{
		reader.setVideoListener(null);
		state = null;
		bufferPool = null;
		bufferQueue = null;
		super.disposeHandler();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.handler.CapabilityHandler#cancelHandler()
	 */
	public void cancelHandler()
	{
		super.cancelHandler();
		try
		{
			releaseBuffer(null);
		}
		catch (InterruptedException e)
		{
		}
		onVideoSetBlock(null);
		synchronized (state)
		{
			state.notify();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.protocol.VideoListener#onVideoConfig(int, int, int,
	 *      int, int)
	 */
	public void onVideoConfig(int frameWidth, int pixelFormat, int frameRate)
	{
		// DEBUG System.out.println("rx VideoConfig");
		synchronized (state)
		{
			this.frameWidth = frameWidth;
			this.frameHeight = frameWidth * 3 / 4;
			// this.frameRate = frameRate;
			this.pixelFormat = pixelFormat;
			state.id = State.READY;
			state.notify();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.protocol.VideoListener#onVideoStart(int)
	 */
	public void onVideoStart(int bufferLength)
	{
		// DEBUG System.out.println("rx VideoStart " + bufferLength);
		synchronized (state)
		{
			state.id = State.ACTIVE;
			state.notify();
			// create a pool of free buffers and a queue for ready buffers
			synchronized (synchQueue)
			{
				bufferQueue.ensureCapacity(poolSize);
			}
			synchronized (synchPool)
			{
				while (poolSize-- > 0)
					bufferPool.push(new byte[bufferLength]);
			}
			this.bufferLength = bufferLength;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.protocol.VideoListener#onVideoStop()
	 */
	public void onVideoStop()
	{
		// DEBUG System.out.println("rx VideoStop");
		synchronized (state)
		{
			state.id = State.INACTIVE;
			state.notify();
			// release buffers
			synchronized (synchPool)
			{
				// at this time no requests for the pool can be pending
				// as the reply to Stop is sent after the last video block
				// has been transmitted
				bufferPool.removeAllElements();
			}
			synchronized (synchQueue)
			{
				// client may be waiting, no goods for it, but wake it up
				bufferQueue.removeAllElements();
				synchQueue.notify();
			}
		}
	}

	/**
	 * Configures video streaming format. Caller is blocked until completion.
	 * <p>
	 * Video resolutions supported go from 80x60 to 160x120 pixels, in steps of
	 * 8x6.
	 * 
	 * @param frameWidth
	 *            The video frame width in pixels (must be multiple of 8)
	 * @param frameHeight
	 *            The video frame height in pixels (must be multiple of 6)
	 * @param pixelFormat
	 *            The identifier of the pixel format
	 * @param frameRate
	 *            (Reserved) Suggested maximum frame rate
	 * 
	 * @throws IOException
	 *             In case the transport layer fails
	 * @throws InterruptedException
	 *             If communication is interrupted before completion
	 * @see VideoSettings
	 */
	public synchronized void configureVideo(int frameWidth, int frameHeight,
		int pixelFormat, int frameRate) throws IOException,
		InterruptedException
	{
		if (cancelled)
			throw new InterruptedException("Operation cancelled");

		if ((frameWidth & 7) != 0 || (frameHeight * 8 != frameWidth * 6))
			throw new IllegalArgumentException(
				"Video frame size must be multiple of 8x6");

		synchronized (state)
		{
			if (state.id == State.ACTIVE)
				throw new IllegalArgumentException(
					"Video streaming must not be active during configuration");
			state.id = State.IDLE;
			writer.sendConfigVideo(frameWidth, pixelFormat, frameRate);
			state.wait();
		}

		if (cancelled)
			throw new InterruptedException("Operation cancelled");
	}

	/**
	 * Starts video streaming. Caller is blocked until completion. Video blocks
	 * can only arrive afterwards.
	 * 
	 * @param bufferLength
	 *            Video block size, zero for automatic settings
	 * @param poolSize
	 *            Number of buffers to allocate
	 * 
	 * @throws IOException
	 *             In case the transport layer fails
	 * @throws InterruptedException
	 *             If communication is interrupted before completion
	 */
	public synchronized void startVideo(int bufferLength, int poolSize)
		throws IOException, InterruptedException
	{
		if (cancelled)
			throw new InterruptedException("Operation cancelled");

		if (poolSize == 0)
			poolSize = 4;

		synchronized (state)
		{
			if (state.id != State.READY)
				throw new IllegalArgumentException(
					"Video streaming must be configured before activation");
			this.poolSize = poolSize;
			writer.sendStartVideo(bufferLength);
			state.wait();
		}

		if (cancelled)
			throw new InterruptedException("Operation cancelled");
	}

	/**
	 * Stops video streaming. Caller is blocked until completion. Video blocks
	 * may still arrive (and should be fetched) during this call, but not
	 * afterwards.
	 * 
	 * @throws IOException
	 *             In case the transport layer fails
	 * @throws InterruptedException
	 *             If communication is interrupted before completion
	 */
	public synchronized void stopVideo() throws IOException,
		InterruptedException
	{
		if (cancelled)
			throw new InterruptedException("Operation cancelled");

		synchronized (state)
		{
			if (state.id != State.ACTIVE)
				throw new IllegalArgumentException(
					"Video streaming must be active before stopping");
			writer.sendStopVideo();
			state.wait();
			state.id = State.READY;
		}

		if (cancelled)
			throw new InterruptedException("Operation cancelled");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.protocol.VideoListener#onVideoGetBlock()
	 */
	public byte[] onVideoGetBlock()
	{
		byte ret[];
		synchronized (synchPool)
		{
			if (bufferPool.size() == 0)
				try
				{
					synchPool.wait();
				}
				catch (InterruptedException e)
				{
					return null;
				}
			ret = (byte[]) bufferPool.pop();
		}
		return ret;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see communication.protocol.VideoListener#onVideoSetBlock(byte[])
	 */
	public void onVideoSetBlock(byte[] buffer)
	{
		synchronized (synchQueue)
		{
			bufferQueue.addElement(buffer);
			synchQueue.notify();
		}
	}

	/**
	 * Retrieves the next video block and acquires ownership. The buffer must be
	 * returned to the pool after use.
	 * 
	 * @return A buffer filled with video data. It may be null, when no more
	 *         data is available (streaming stopped).
	 * 
	 * @throws InterruptedException
	 *             If communication is interrupted before completion
	 * @see #releaseBuffer(byte[])
	 */
	public byte[] acquireBuffer() throws InterruptedException
	{
		if (cancelled)
			throw new InterruptedException("Operation cancelled");

		synchronized (state)
		{
			if (state.id != State.ACTIVE)
				return null;
		}
		byte[] ret = null;
		synchronized (synchQueue)
		{
			if (bufferQueue.size() == 0)
				try
				{
					synchQueue.wait();
				}
				catch (InterruptedException e)
				{
					return null;
				}
			if (cancelled)
				throw new InterruptedException("Operation cancelled");
			if (bufferQueue.size() > 0)
			{
				ret = (byte[]) bufferQueue.firstElement();
				bufferQueue.removeElementAt(0);
			}
		}
		return ret;
	}

	/**
	 * Releases ownership of a video block after processing and recycles it for
	 * later use.
	 * 
	 * @param buffer
	 *            A buffer that has been obtained by a previous call to
	 *            {@link #acquireBuffer()}
	 * 
	 * @throws InterruptedException
	 *             If communication is interrupted before completion
	 */
	public void releaseBuffer(byte[] buffer) throws InterruptedException
	{
		if (cancelled)
			throw new InterruptedException("Operation cancelled");

		synchronized (synchPool)
		{
			bufferPool.push(buffer);
			synchPool.notify();
		}

		if (cancelled)
			throw new InterruptedException("Operation cancelled");
	}

	/**
	 * Obtains the configured frame height
	 * 
	 * @return The frame height in pixels
	 * 
	 * @see #configureVideo
	 * @see VideoSettings#HEIGHT_DEFAULT
	 * @see VideoSettings#HEIGHT_MAX
	 */
	public int getFrameHeight()
	{
		return frameHeight;
	}

	/**
	 * Obtains the configured frame width
	 * 
	 * @return The frame width in pixels
	 * 
	 * @see #configureVideo
	 * @see VideoSettings#WIDTH_DEFAULT
	 * @see VideoSettings#WIDTH_MAX
	 */
	public int getFrameWidth()
	{
		return frameWidth;
	}

	/**
	 * Obtains the configured pixel format
	 * 
	 * @return The pixel format identifier
	 * 
	 * @see #configureVideo
	 * @see VideoSettings#FMT_RGB656
	 * @see VideoSettings#FMT_RGB656_SMOOTH
	 */
	public int getPixelFormat()
	{
		return pixelFormat;
	}

	/**
	 * Obtains the approved video buffers length
	 * 
	 * @return The accepted buffer length in bytes
	 * 
	 * @see #startVideo
	 */
	public int getBufferLength()
	{
		return bufferLength;
	}
}
