package communication.protocol;

import java.io.IOException;
import java.io.InputStream;

/**
 * Use to receive and decode protocol chunks, and call the appropriate protocol
 * listeners.
 * <p>
 * Chunks are extracted from an InputStream, as obtained from a
 * {@link communication.transport.ConnectionProvider}
 * 
 * @author Paolo
 */
public class ChunkReader
{
	private volatile InputStream input;

	private final byte buf1[] = new byte[1];

	private final byte buf2[] = new byte[2];

	private final byte buf3[] = new byte[3];

	private ProtocolListener protocolListener;

	private InternalListener internalListener;

	private VideoListener videoListener;

	private TransferListener transferListener;

	private BusListener busListener;

	private volatile boolean cancelled;

	private boolean useAvailable;

	/**
	 * Sets the listener for received chunks in the Protocol capability
	 * 
	 * @param listener
	 *            The new listener to set
	 */
	public void setProtocolListener(ProtocolListener listener)
	{
		protocolListener = listener;
	}

	/**
	 * Sets the listener for received chunks in the Internal capability
	 * 
	 * @param listener
	 *            The new listener to set
	 */
	public void setInternalListener(InternalListener listener)
	{
		internalListener = listener;
	}

	/**
	 * Sets the listener for received chunks in the Video capability
	 * 
	 * @param listener
	 *            The new listener to set
	 */
	public void setVideoListener(VideoListener listener)
	{
		videoListener = listener;
	}

	/**
	 * Sets the listener for received chunks in the Transfer capability
	 * 
	 * @param listener
	 *            The new listener to set
	 */
	public void setTransferListener(TransferListener listener)
	{
		transferListener = listener;
	}

	/**
	 * Sets the listener for received chunks in the Bus capability
	 * 
	 * @param listener
	 *            The new listener to set
	 */
	public void setBusListener(BusListener listener)
	{
		busListener = listener;
	}

	static private int unsigned(byte val)
	{
		return (((int) val) & 0xFF);
	}

	static private int unsigned(byte low, byte high)
	{
		return (((int) low) & 0xFF) | ((((int) high) & 0xFF) << 8);
	}

	private int readByte() throws IOException
	{
		if (cancelled)
			return -1;
		int r;
		if (useAvailable)
			while (input.available() < 1)
			{
				if (cancelled)
					return -1;
			}
		do
		{
			if (cancelled)
				return -1;
			r = input.read();
		}
		while (r < 0);
		return r;
	}

	private boolean readBytes(byte buf[]) throws IOException
	{
		if (cancelled)
			return false;
		if (useAvailable)
			while (input.available() < buf.length)
			{
				if (cancelled)
					return false;
			}
		int c = 0;
		while (c < buf.length)
		{
			if (cancelled)
				return false;
			int r = input.read(buf, c, buf.length - c);
			if (r > 0)
				c += r;
		}
		return true;
	}

	/**
	 * Releases the input stream.
	 * 
	 * @return The owned input stream
	 */
	public InputStream detachInput()
	{
		InputStream ret = input;
		input = null;
		cancelled = true;
		return ret;
	}

	/**
	 * Acquires an input stream.
	 * 
	 * @param input
	 *            The input stream on which to gain ownership
	 * @param useAvailable
	 *            Whether InputStream.available() should be used to simulate
	 *            non-blocking read().
	 * 
	 * @see communication.transport.ConnectionProvider#getStreamProperties()
	 */
	public void attachInput(InputStream input, boolean useAvailable)
	{
		cancelled = false;
		this.input = input;
		this.useAvailable = useAvailable;
	}

	/**
	 * Cancels any pending read operation, if possible.
	 * 
	 * @see #receiveChunk()
	 */
	public void cancel()
	{
		cancelled = true;
	}

	/**
	 * Reads a chunk from the input stream and call the appropriate listener.
	 * This method blocks until the whole chunk has been received and the
	 * listener returned. It can be cancelled from another thread by calling
	 * cancel().
	 * 
	 * @return The method returns true if completed successfully, false if
	 *         cancelled before completion.
	 * 
	 * @see #cancel()
	 * 
	 * @throws IOException
	 */
	synchronized public boolean receiveChunk() throws IOException
	{
		int chunk = readByte();
		if (chunk < 0)
			return false;
		if (protocolListener != null)
			protocolListener.onChunk(chunk);

		switch (chunk)
		{
		// Basic protocol packets

		case Chunk.PING:
			if (protocolListener != null)
				protocolListener.onPing();
			break;
		case Chunk.PONG:
			if (protocolListener != null)
				protocolListener.onPong();
			break;
		case Chunk.SETUP:
			if (!readBytes(buf2))
				return false;
			if (protocolListener != null)
				protocolListener.onSetup(unsigned(buf2[0]),
					9600 * unsigned(buf2[1]));
			break;
		case Chunk.DISCONNECT:
			if (protocolListener != null)
				protocolListener.onDisconnect();
			return false;

		// Video Capability

		case Chunk.VIDEO_CONFIG:
			if (!readBytes(buf3))
				return false;
			if (videoListener != null)
				videoListener.onVideoConfig(8 * unsigned(buf3[0]),
					unsigned(buf3[1]), buf3[2]);
			break;
		case Chunk.VIDEO_START:
			if (!readBytes(buf2))
				return false;
			if (videoListener != null)
				videoListener.onVideoStart(unsigned(buf2[0], buf2[1]));
			break;
		case Chunk.VIDEO_STOP:
			if (videoListener != null)
				videoListener.onVideoStop();
			break;
		case Chunk.VIDEO_BLOCK:
			if (videoListener != null)
			{
				byte buf[] = videoListener.onVideoGetBlock();
				if (buf != null)
				{
					if (!readBytes(buf))
						return false;
					videoListener.onVideoSetBlock(buf);
				}
			}
			break;

		// Internal Capability

		case Chunk.INTERNAL_DATA1:
			if (!readBytes(buf1))
				return false;
			if (internalListener != null)
				internalListener.onInternalRegData1(unsigned(buf1[0]));
			break;
		case Chunk.INTERNAL_DATA2:
			if (!readBytes(buf2))
				return false;
			if (internalListener != null)
				internalListener.onInternalRegData2(unsigned(buf2[0]),
					unsigned(buf2[1]));
			break;
		case Chunk.INTERNAL_DATA3:
			if (!readBytes(buf3))
				return false;
			if (internalListener != null)
				internalListener.onInternalRegData3(unsigned(buf3[0]),
					unsigned(buf3[1]), unsigned(buf3[2]));
			break;
		case Chunk.INTERNAL_READERR:
			if (internalListener != null)
				internalListener.onInternalReadError();
			break;
		case Chunk.INTERNAL_WRITEERR:
			if (internalListener != null)
				internalListener.onInternalWriteError();
			break;
		case Chunk.INTERNAL_WRITEOK:
			if (internalListener != null)
				internalListener.onInternalWriteOk();
			break;

		// Bus Capability

		case Chunk.BUS_RESULT:
			int n = readByte();
			if (n < 0)
				return false;
			byte bdata[][] = null;
			if (n < 255) // not error
			{
				bdata = new byte[n][];
				for (int i = 0; i < n; ++i)
				{
					int m = readByte();
					if (m < 0)
						return false;
					bdata[i] = new byte[m];
					if (!readBytes(bdata[i]))
						return false;
				}
			}
			if (busListener != null)
				busListener.onBusResult(bdata);
			bdata = null; // help GC
			break;

		// Transfer Capability

		case Chunk.FILE_RESULT:
			if (!readBytes(buf2))
				return false;
			if (transferListener != null)
				transferListener.onTransferResult(unsigned(buf2[0]),
					unsigned(buf2[1]));
			break;
		case Chunk.FILE_DATA:
			if (!readBytes(buf3))
				return false;
			int size = unsigned(buf3[1], buf3[2]);
			byte[] tdata = new byte[size];
			if (!readBytes(tdata))
				return false;
			if (transferListener != null)
				transferListener.onTransferData(unsigned(buf3[0]), tdata);
			tdata = null; // help GC
			break;
		case Chunk.DIR_ENTRY:
			if (!readBytes(buf3))
				return false;
			byte[] ddata = new byte[unsigned(buf3[2])];
			if (!readBytes(ddata))
				return false;
			if (transferListener != null)
				transferListener.onDirectoryEntry(unsigned(buf3[0]),
					new String(ddata), unsigned(buf3[1]));
			ddata = null; // help GC
			break;
		}
		return true;
	}
}
