package communication.protocol;

import java.io.IOException;
import java.io.OutputStream;

/**
 * Use to encode and send protocol chunks.
 * <p>
 * Chunks are inserted to an OutputStream, as obtained from a
 * {@link communication.transport.ConnectionProvider}
 * 
 * @author Paolo
 */
public class ChunkWriter
{
	private volatile OutputStream output;

	private static int regaddr(int module, int register)
	{
		return ((module & 7) << 5) | (register & 31);
	}

	private static byte[] word2bytes(int word)
	{
		final byte[] buf2 = new byte[2];
		buf2[0] = (byte) (word & 255);
		buf2[1] = (byte) (word >> 8);
		return buf2;
	}

	/**
	 * Releases the output stream.
	 * 
	 * @return The owned output stream
	 */
	public OutputStream detachOutput()
	{
		OutputStream ret = output;
		output = null;
		return ret;
	}

	/**
	 * Acquires an output stream.
	 * 
	 * @param output
	 *            The output stream on which to gain ownership
	 */
	public void attachOutput(OutputStream output)
	{
		this.output = output;
	}

	// Basic protocol

	/**
	 * Receives confirmation of communication setup and approved settings.
	 * 
	 * @param capabilityMask
	 *            Requested capabilities, a combination of values defined in
	 *            {@link Capability}
	 * @param maxSpeed
	 *            Maximum bitrate at UART level (mainly statistical)
	 * 
	 * @throws IOException
	 *             Propagated from the underlying output stream
	 * @see Chunk#SETUP
	 */
	synchronized public void sendSetup(int capabilityMask, int maxSpeed)
		throws IOException
	{
		output.write(Chunk.SETUP);
		output.write(capabilityMask);
		output.write(maxSpeed / 9600);
		output.flush();
	}

	/**
	 * Sends a ping request
	 * 
	 * @throws IOException
	 *             Propagated from the underlying output stream
	 * @see Chunk#PING
	 */
	synchronized public void sendPing() throws IOException
	{
		output.write(Chunk.PING);
		output.flush();
	}

	/**
	 * Sends a ping reply
	 * 
	 * @throws IOException
	 *             Propagated from the underlying output stream
	 * @see Chunk#PING
	 */
	synchronized public void sendPong() throws IOException
	{
		output.write(Chunk.PONG);
		output.flush();
	}

	/**
	 * Sends a disconnection indication
	 * 
	 * @throws IOException
	 *             Propagated from the underlying output stream
	 * @see Chunk#DISCONNECT
	 */
	synchronized public void sendDisconnect() throws IOException
	{
		output.write(Chunk.DISCONNECT);
		output.flush();
	}

	// Internal Capability

	/**
	 * Sends request of reading 1 byte from internal registers
	 * 
	 * @param module
	 *            Internal module identifier (0-7)
	 * @param register
	 *            Module register (in the range 0-31)
	 * @throws IOException
	 *             Propagated from the underlying output stream
	 * @see Chunk#INTERNAL_READ1
	 */
	synchronized public void sendReadInternalReg1(int module, int register)
		throws IOException
	{
		output.write(Chunk.INTERNAL_READ1);
		output.write(regaddr(module, register));
		output.flush();
	}

	/**
	 * Sends request of reading 2 bytes from internal registers
	 * 
	 * @param module
	 *            Internal module identifier (0-7)
	 * @param register
	 *            Module register (in the range 0-31)
	 * @throws IOException
	 *             Propagated from the underlying output stream
	 * @see Chunk#INTERNAL_READ2
	 */
	synchronized public void sendReadInternalReg2(int module, int register)
		throws IOException
	{
		output.write(Chunk.INTERNAL_READ2);
		output.write(regaddr(module, register));
		output.flush();
	}

	/**
	 * Sends request of reading 3 bytes from internal registers
	 * 
	 * @param module
	 *            Internal module identifier (0-7)
	 * @param register
	 *            Module register (in the range 0-31)
	 * @throws IOException
	 *             Propagated from the underlying output stream
	 * @see Chunk#INTERNAL_READ3
	 */
	synchronized public void sendReadInternalReg3(int module, int register)
		throws IOException
	{
		output.write(Chunk.INTERNAL_READ3);
		output.write(regaddr(module, register));
		output.flush();
	}

	/**
	 * Sends request of writing 1 byte to internal registers
	 * 
	 * @param module
	 *            Internal module identifier (0-7)
	 * @param register
	 *            Module register (in the range 0-31)
	 * @param value
	 *            Data for the specified register
	 * @throws IOException
	 *             Propagated from the underlying output stream
	 * @see Chunk#INTERNAL_WRITE1
	 */
	synchronized public void sendWriteInternalReg1(int module, int register,
		int value) throws IOException
	{
		output.write(Chunk.INTERNAL_WRITE1);
		output.write(regaddr(module, register));
		output.write(value);
		output.flush();
	}

	/**
	 * Sends request of writing 2 bytes to internal registers
	 * 
	 * @param module
	 *            Internal module identifier (0-7)
	 * @param register
	 *            Module register (in the range 0-31)
	 * @param value1
	 *            Data for the first register
	 * @param value2
	 *            Data for the second register (or the MSB in a 16-bit register)
	 * @throws IOException
	 *             Propagated from the underlying output stream
	 * @see Chunk#INTERNAL_WRITE2
	 */
	synchronized public void sendWriteInternalReg2(int module, int register,
		int value1, int value2) throws IOException
	{
		output.write(Chunk.INTERNAL_WRITE2);
		output.write(regaddr(module, register));
		output.write(value1);
		output.write(value2);
		output.flush();
	}

	/**
	 * Sends request of writing 3 bytes to internal registers
	 * 
	 * @param module
	 *            Internal module identifier (0-7)
	 * @param register
	 *            Module register (in the range 0-31)
	 * @param value1
	 *            Data for the first register
	 * @param value2
	 *            Data for the second register (or the MSB in a 16-bit register)
	 * @param value3
	 *            Data for the third register (or the following byte in previous
	 *            registers)
	 * @throws IOException
	 *             Propagated from the underlying output stream
	 * @see Chunk#INTERNAL_WRITE3
	 */
	synchronized public void sendWriteInternalReg3(int module, int register,
		int value1, int value2, int value3) throws IOException
	{
		output.write(Chunk.INTERNAL_WRITE3);
		output.write(regaddr(module, register));
		output.write(value1);
		output.write(value2);
		output.write(value3);
		output.flush();
	}

	// Bus Capability

	/**
	 * Sends a bus transfer request
	 * 
	 * @param addr
	 *            Array of I2C slave devices each message is addressed to
	 * @param data
	 *            Array of buffers containing data to transfer for write
	 *            messages or a single byte count for read messages
	 * @throws IOException
	 *             Propagated from the underlying output stream
	 * @see Chunk#BUS_TRANSFER
	 */
	synchronized public void sendBusTransfer(int[] addr, boolean[] readFlag,
		byte[][] data) throws IOException
	{
		output.write(Chunk.BUS_TRANSFER);
		int n = addr.length & 255;
		output.write(n);
		for (int i = 0; i < n; ++i)
		{
			output.write(addr[i] & 0x7F | (readFlag[i] ? 0x80 : 0));
			int m = readFlag[i] ? ((int) data[i][0] & 255)
				: data[i].length & 255;
			output.write(m);
			if (!readFlag[i])
				output.write(data[i], 0, m);
		}
		output.flush();
	}

	/**
	 * Sends a delay between bus transfers
	 * 
	 * @param millisecs
	 *            The pause duration in milliseconds
	 * @throws IOException
	 *             Propagated from the underlying output stream
	 * @see Chunk#BUS_PAUSE
	 */
	synchronized public void sendBusPause(int millisecs) throws IOException
	{
		output.write(Chunk.BUS_PAUSE);
		output.write(word2bytes(millisecs));
	}

	// Video Capability

	/**
	 * Sends request for configuring video stream parameters.
	 * 
	 * @param frameWidth
	 *            The requested video frame width (must be multiple of 8). Frame
	 *            height can be obtained by aspect ratio of 4:3
	 * @param pixelFormat
	 *            The identifier of the pixel format
	 * @param frameRate
	 *            (Reserved) Suggested maximum frame rate
	 * 
	 * @throws IOException
	 *             Propagated from the underlying output stream
	 * @see Chunk#VIDEO_CONFIG
	 */
	synchronized public void sendConfigVideo(int frameWidth, int pixelFormat,
		int frameRate) throws IOException
	{
		output.write(Chunk.VIDEO_CONFIG);
		output.write(frameWidth / 8);
		output.write(pixelFormat);
		output.write(frameRate);
		output.flush();
	}

	/**
	 * Sends request of starting video streaming with the specified block size.
	 * 
	 * @param blockLength
	 *            The requested size of video blocks
	 * 
	 * @throws IOException
	 *             Propagated from the underlying output stream
	 * @see Chunk#VIDEO_START
	 */
	synchronized public void sendStartVideo(int blockLength) throws IOException
	{
		output.write(Chunk.VIDEO_START);
		output.write(word2bytes(blockLength));
		output.flush();
	}

	/**
	 * Sends request to terminate video streaming.
	 * 
	 * @throws IOException
	 *             Propagated from the underlying output stream
	 * @see Chunk#VIDEO_STOP
	 */
	synchronized public void sendStopVideo() throws IOException
	{
		output.write(Chunk.VIDEO_STOP);
		output.flush();
	}

	// Transfer Capability

	/**
	 * Sends command to open/create a remote file
	 * 
	 * @param flags
	 *            Specify 1 to read an existing file, 2 to create a new file in
	 *            write mode
	 * @param name
	 *            Path to the remote file. Use a relative path to access files
	 *            in the RW flash partition, absolute path to access files
	 *            anywhere
	 * @throws IOException
	 *             Propagated from the underlying output stream
	 * @see Chunk#FILE_OPEN
	 * @see Chunk#FILE_RESULT
	 */
	synchronized public void sendFileOpen(int flags, String name)
		throws IOException
	{
		output.write(Chunk.FILE_OPEN);
		output.write(flags);
		output.write(name.length());
		output.write(name.getBytes()); // TODO "ascii-us"
		output.flush();
	}

	/**
	 * Sends command to close a remote file
	 * 
	 * @param token
	 *            The identifier of an open remote file
	 * @throws IOException
	 *             Propagated from the underlying output stream
	 * @see Chunk#FILE_CLOSE
	 * @see Chunk#FILE_RESULT
	 */
	synchronized public void sendFileClose(int token) throws IOException
	{
		output.write(Chunk.FILE_CLOSE);
		output.write(token);
		output.flush();
	}

	/**
	 * Sends command to write data to a remote file
	 * 
	 * @param token
	 *            The identifier of an open remote file
	 * @param data
	 *            A byte array containing raw data to write
	 * @throws IOException
	 *             Propagated from the underlying output stream
	 * @see Chunk#FILE_WRITE
	 * @see Chunk#FILE_RESULT
	 */
	synchronized public void sendFileWrite(int token, byte[] data)
		throws IOException
	{
		output.write(Chunk.FILE_WRITE);
		output.write(token);
		output.write(word2bytes(data.length));
		output.write(data);
		output.flush();
	}

	/**
	 * Sends request to read data from a remote file
	 * 
	 * @param token
	 *            The identifier of an open remote file
	 * @param size
	 *            The length of raw data to read
	 * @throws IOException
	 *             Propagated from the underlying output stream
	 * @see Chunk#FILE_READ
	 * @see Chunk#FILE_RESULT
	 * @see Chunk#FILE_DATA
	 */
	synchronized public void sendFileRead(int token, int size)
		throws IOException
	{
		output.write(Chunk.FILE_READ);
		output.write(token);
		output.write(word2bytes(size));
		output.flush();
	}

	/**
	 * Sends command to close a remote file
	 * 
	 * @param token
	 *            The identifier of an open remote file
	 * @throws IOException
	 *             Propagated from the underlying output stream
	 * @see Chunk#FILE_RESULT
	 */
	synchronized public void sendFileFlush(int token) throws IOException
	{
		output.write(Chunk.FILE_FLUSH);
		output.write(token);
		output.flush();
	}

	/**
	 * Sends request to read the next entry in a remote directory
	 * 
	 * @param token
	 *            The identifier of an open remote file
	 * @throws IOException
	 *             Propagated from the underlying output stream
	 * @see Chunk#DIR_NEXT
	 * @see Chunk#DIR_ENTRY
	 * @see Chunk#FILE_RESULT
	 */
	synchronized public void sendDirectoryNext(int token) throws IOException
	{
		output.write(Chunk.DIR_NEXT);
		output.write(token);
		output.flush();
	}

	// TODO transfer server interface: result, data
}
